# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class Robot(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsRobot(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = Robot()
        x.Init(buf, n + offset)
        return x

    # Robot
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # Robot
    def Id(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Robot
    def Immovable(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(self._tab.Get(tdw.flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Robot
    def Transform(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from .SimpleTransform import SimpleTransform
            obj = SimpleTransform()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Robot
    def Joints(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = self._tab.Vector(o)
            x += tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from .RobotJoint import RobotJoint
            obj = RobotJoint()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Robot
    def JointsLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def RobotStart(builder): builder.StartObject(4)
def RobotAddId(builder, id): builder.PrependInt32Slot(0, id, 0)
def RobotAddImmovable(builder, immovable): builder.PrependBoolSlot(1, immovable, 0)
def RobotAddTransform(builder, transform): builder.PrependStructSlot(2, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(transform), 0)
def RobotAddJoints(builder, joints): builder.PrependUOffsetTRelativeSlot(3, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(joints), 0)
def RobotStartJointsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def RobotEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class Environments(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsEnvironments(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = Environments()
        x.Init(buf, n + offset)
        return x

    # Environments
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # Environments
    def Envs(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j) * 28
            from .Environment import Environment
            obj = Environment()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Environments
    def EnvsLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

def EnvironmentsStart(builder): builder.StartObject(1)
def EnvironmentsAddEnvs(builder, envs): builder.PrependUOffsetTRelativeSlot(0, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(envs), 0)
def EnvironmentsStartEnvsVector(builder, numElems): return builder.StartVector(28, numElems, 4)
def EnvironmentsEnd(builder): return builder.EndObject()



from os import urandom
import numpy as np
import math
import base64
import json
import scipy.signal as sg
from enum import Enum
from typing import Dict, Optional, Tuple, Union, List
from tdw.output_data import OutputData, Rigidbodies, Collision, EnvironmentCollision
from pathlib import Path
from pkg_resources import resource_filename
from csv import DictReader
import io


class AudioMaterial(Enum):
    """
    These are the materials currently supported for impact sounds in pyImpact.  More will be added in time.
    """

    ceramic = 0
    glass = 1
    metal = 2
    wood_hard = 3
    wood_medium = 4
    wood_soft = 5
    cardboard = 6
    paper = 7
    plastic_hard = 8
    plastic_soft_foam = 9
    rubber = 10
    fabric = 11
    leather = 12
    stone = 13


class ObjectInfo:
    """
    Impact sound data for an object in a TDW model library.
    The audio values here are just recommendations; you can apply different values if you want.
    """

    def __init__(self, name: str, amp: float, mass: float, material: AudioMaterial, library: str, bounciness: float, resonance: float, size: int):
        """
        :param name: The model name.
        :param amp: The sound amplitude.
        :param mass: The object mass.
        :param material: The audio material.
        :param library: The path to the model library (see ModelLibrarian documentation).
        :param bounciness: The bounciness value for a Unity physics material.
        :param resonance: The resonance value for the object.
        :param size: Integer representing the size "bucket" this object belongs to (0-5).
        """

        self.amp = amp
        self.library = library
        self.mass = mass
        self.material = material
        self.name = name
        self.bounciness = bounciness
        self.resonance = resonance
        self.size = size


# Density per audio material.
DENSITIES: Dict[AudioMaterial, float] = {AudioMaterial.ceramic: 2180,
                                         AudioMaterial.glass: 2500,
                                         AudioMaterial.stone: 2000,
                                         AudioMaterial.metal: 8450,
                                         AudioMaterial.wood_hard: 1200,
                                         AudioMaterial.wood_medium: 700,
                                         AudioMaterial.wood_soft: 400,
                                         AudioMaterial.fabric: 1540,
                                         AudioMaterial.leather: 860,
                                         AudioMaterial.plastic_hard: 1150,
                                         AudioMaterial.plastic_soft_foam: 285,
                                         AudioMaterial.rubber: 1522,
                                         AudioMaterial.paper: 1200,
                                         AudioMaterial.cardboard: 698}


class Base64Sound:
    """
    A sound encoded as a base64 string.

    Fields:

    - `bytes` Byte data of the sound.
    - `wav_str` base64 string of the sound. Send this to the build.
    - `length` The length of the byte array.
    """

    def __init__(self, snd: np.array):
        """
        :param snd: The sound byte array.
        """

        tst1 = np.array(snd * 32767, dtype='int16')
        tst2 = bytes(tst1)
        tst3 = base64.b64encode(tst2).decode('utf-8')

        self.bytes = tst2
        self.wav_str = tst3
        self.length = len(tst2)


class Modes:
    """
    Resonant mode properties: Frequencies, powers, and times.
    """

    def __init__(self, frequencies: np.array, powers: np.array, decay_times: np.array):
        """
        :param frequencies: numpy array of mode frequencies in Hz
        :param powers: numpy array of mode onset powers in dB re 1.
        :param decay_times: numpy array of mode decay times (i.e. the time in ms it takes for each mode to decay 60dB from its onset power)
        """
        self.frequencies = frequencies
        self.powers = powers
        self.decay_times = decay_times

    def sum_modes(self, fs: int = 44100, resonance: float = 1.0) -> np.array:
        """
        Create mode time-series from mode properties and sum them together.

        :return A synthesized sound.
        """

        # Scroll through modes
        for i in range(len(self.frequencies)):
            H_dB = 80 + self.powers[i]
            L_ms = self.decay_times[i] * H_dB / 60
            mLen = math.ceil(L_ms / 1e3 * fs)
            # if mode length is greater than the current time-series we have had make our time series longer
            max_len = mLen
            if mLen > max_len:
                max_len = mLen
            tt = np.arange(0, max_len) / fs
            # synthesize a sinusoid
            mode = np.cos(2 * math.pi * self.frequencies[i] * tt)
            mode = mode * (10 ** (self.powers[i] / 20))
            dcy = tt * (60 / (self.decay_times[i] * resonance / 1e3))
            env = 10 ** (-dcy / 20)
            mode = mode * env
            if i == 0:
                synth_sound = mode
            else:
                synth_sound = Modes.mode_add(synth_sound, mode)
        return synth_sound

    @staticmethod
    def mode_add(a: np.array, b: np.array) -> np.array:
        """
        Add together numpy arrays of different lengths by zero-padding the shorter.

        :param a: The first array.
        :param b: The second array.

        :return The summed modes.
        """

        if len(a) < len(b):
            c = b.copy()
            c[:len(a)] += a
        else:
            c = a.copy()
            c[:len(b)] += b
        return c


class CollisionInfo:
    """
    Class containing information about collisions required by pyImpact to determine the volume of impact sounds.
    """

    def __init__(self, obj1_modes: Modes, obj2_modes: Modes, amp: float = 0.5, init_speed: float = 1):
        """
        :param amp: Amplitude of the first collision (must be between 0 and 1).
        :param init_speed: The speed of the initial collision (all collisions will be scaled relative to this).
        :param obj1_modes: The object's modes.
        :param obj2_modes: The other object's modes.
        """

        self.count = 0
        self.amp = amp
        # The speed of the initial collision.
        self.init_speed = init_speed
        # The audio modes.
        self.obj1_modes = obj1_modes
        self.obj2_modes = obj2_modes

    def count_collisions(self) -> None:
        """
        Update the counter for how many times two objects have collided.
        """

        self.count += 1


class CollisionType(Enum):
    """
    The "type" of a collision, defined by the motion of the object.

    none = No collision
    impact = The object "entered" a collision
    scrape = The object "stayed" in a collision with a low angular velocity.
    roll = The object "stayed" in a collision with a high angular velocity.
    """

    none = 1,
    impact = 2,
    scrape = 4,
    roll = 8


class CollisionTypesOnFrame:
    """
    All types of collision (impact, scrape, roll, none) between an object and any other objects or the environment on this frame.

    Usage:

    ```python
    from tdw.controller import Controller
    from tdw.py_impact import CollisionTypesOnFrame

    object_id = c.get_unique_id()
    c = Controller()
    c.start()

    # Your code here.

    # Request the required output data (do this at the start of the simulation, not per frame).
    resp = c.communicate([{"$type": "send_collisions",
                           "enter": True,
                           "exit": False,
                           "stay": True,
                           "collision_types": ["obj", "env"]},
                          {"$type": "send_rigidbodies",
                           "frequency": "always"}])

    # Parse the output data and get collision type data.
    ctof = CollisionTypesOnFrame(object_id, resp)

    # Read the dictionaries of collidee IDs and collision types.
    for collidee_id in ctof.collisions:
        collision_type = ctof.collisions[collidee_id]
        print(collidee_id, collision_type)

    # Check the environment collision.
    print(ctof.env_collision_type)
    ```
    """

    def __init__(self, object_id: int, resp: List[bytes]):
        """
        :param object_id: The unique ID of the colliding object.
        :param resp: The response from the build.
        """

        collisions, env_collisions, rigidbodies = PyImpact.get_collisions(resp)
        # The type of collision with each collidee.
        self.collisions: Dict[int, CollisionType] = dict()
        # The type of environment collision, if any.
        self.env_collision = CollisionType.none

        # If there is no Rigidbodies output data, then nothing can be parsed.
        if rigidbodies is None:
            return

        # Get the rigidbody data for this object.
        for i in range(rigidbodies.get_num()):
            if rigidbodies.get_id(i) == object_id:
                # Get the angular velocity of this object.
                ang_vel = rigidbodies.get_angular_velocity(i)

                # My collisions with other objects on this frame.
                # Key = the collidee ID. Value = list of states.
                my_collisions: Dict[int, List[str]] = dict()

                for co in collisions:
                    if co.get_collider_id() == object_id:
                        collidee = co.get_collidee_id()
                        if collidee not in my_collisions:
                            my_collisions.update({collidee: []})
                        my_collisions[collidee].append(co.get_state())
                # Get the collision type.
                for collidee in my_collisions:
                    self.collisions[collidee] = self._get_collision_type(ang_vel, my_collisions[collidee])
                env_collision_states: List[str] = []
                for co in env_collisions:
                    env_collision_states.append(co.get_state())

    @staticmethod
    def _get_collision_type(ang_vel: tuple, states: List[str]) -> CollisionType:
        """
        :param ang_vel: The angular velocity of this object.

        :param states: The states of all collisions experienced by this object on this frame.

        :return: The type of collision that the object is experiencing.
        """

        # If there is any "enter" state in this frame, then it is an impact.
        if "enter" in states:
            return CollisionType.impact
        # If there are "stay" state(s) but no "enter" state, then it is either a roll or a scrape.
        elif "stay" in states:
            # If there is a high angular velocity, then it is a roll.
            if np.linalg.norm(ang_vel) > 1:
                return CollisionType.roll
            # Otherwise, it's a scrape.
            else:
                return CollisionType.scrape
        else:
            return CollisionType.none


class PyImpact:
    """
    Generate impact sounds from physics data.

    Sounds are synthesized as described in: [Traer,Cusimano and McDermott, A PERCEPTUALLY INSPIRED GENERATIVE MODEL OF RIGID-BODY CONTACT SOUNDS, Digital Audio Effects, (DAFx), 2019](http://dafx2019.bcu.ac.uk/papers/DAFx2019_paper_57.pdf)

    For a general guide on impact sounds in TDW, read [this](../misc_frontend/impact_sounds.md).

    For example usage, see: `tdw/Python/example_controllers/impact_sounds.py`
    """

    def __init__(self, initial_amp: float = 0.5, prevent_distortion: bool = True, logging: bool = False):
        """
        :param initial_amp: The initial amplitude, i.e. the "master volume". Must be > 0 and < 1.
        :param prevent_distortion: If True, clamp amp values to <= 0.99
        :param logging: If True, log mode properties for all colliding objects, as json.
        """

        assert 0 < initial_amp < 1, f"initial_amp is {initial_amp} (must be > 0 and < 1)."

        self.initial_amp = initial_amp
        self.prevent_distortion = prevent_distortion
        self.logging = logging

        # The collision info per set of objects.
        self.object_modes: Dict[int, Dict[int, CollisionInfo]] = {}

        # Cache the material data. This is use to reset the material modes.
        self.material_data: Dict[str, dict] = {}
        material_list = ["ceramic", "wood_hard", "wood_medium", "wood_soft", "metal", "glass", "paper", "cardboard", "leather", "fabric", "plastic_hard", "plastic_soft_foam", "rubber", "stone"]
        for mat in material_list:
            for i in range(6):
                # Load the JSON data.
                mat_name = mat + "_" + str(i)
                path = mat_name + "_mm"
                data = json.loads(Path(resource_filename(__name__, f"py_impact/material_data/{path}.json")).read_text())
                self.material_data.update({mat_name: data})

        # Create empty dictionary for log.
        self.mode_properties_log = dict()

        # Get the default object info. See: `set_default_audio_info()`
        self.object_info: Dict[str, ObjectInfo] = dict()
        # A dictionary of objects in the scene. See: `set_default_audio_info()`
        self.object_names: Dict[int, str] = dict()
        # A dummy ID for the environment. See: `set_default_audio_info()`
        self.env_id: int = -1

    def set_default_audio_info(self, object_names: Dict[int, str]) -> None:
        """
        Set the default audio info for each object in the scene.

        :param object_names: A dictionary of objects in the scene. Key = The object ID. Value = The name of the model.
        """

        self.object_names = object_names
        self.object_info = PyImpact.get_object_info()
        self.env_id = int.from_bytes(urandom(3), byteorder='big')

    def get_audio_commands(self, resp: List[bytes], floor: AudioMaterial, wall: AudioMaterial, resonance_audio: bool = False) -> List[dict]:
        """
        Given the response from the build, get audio commands.

        Required:

        - You've already called `set_default_audio_info()` and supplied it with the IDs and names of each object in the scene:
        - You've requested collision and rigidbody data per-frame (`send_collisions` and `send_rigidbodies`)

        :param resp: The response from the build.
        :param floor: The floor material as an `AudioMaterial` enum value.
        :param wall: The wall material as an `AudioMaterial` enum value.
        :param resonance_audio: If True, this function assumes you're using Resonance Audio.

        :return: A list of commands that will play audio generated by the collisions on this frame. This can be empty.
        """

        commands = []
        collisions, env_collisions, rigidbodies = self.get_collisions(resp=resp)
        # Get the mass and speed of each object.
        speeds: Dict[int, float] = dict()
        for i in range(rigidbodies.get_num()):
            object_id = rigidbodies.get_id(i)
            speeds[object_id] = np.linalg.norm(rigidbodies.get_velocity(i))
        # Get all stays and exits.
        stays_and_exits: List[int] = list()
        for c in collisions:
            if c.get_state() == "stay" or c.get_state() == "exit":
                stays_and_exits.extend([c.get_collider_id(), c.get_collidee_id()])
        colliders: List[int] = list()
        # Play sounds from collisions.
        for collision in collisions:
            # Ignore invalid collisions.
            if collision.get_state() != "enter" or not self.is_valid_collision(collision=collision):
                continue
            collider_id = collision.get_collider_id()
            collidee_id = collision.get_collidee_id()
            # Ignore collisions if there is a "stay" event, because this can create a droning effect.
            if collider_id in stays_and_exits or collidee_id in stays_and_exits or\
                    collider_id in colliders or collidee_id in colliders:
                continue
            colliders.append(collider_id)
            colliders.append(collidee_id)
            # Skip objects that for some reason aren't in the cached data.
            if collider_id not in self.object_names or collidee_id not in self.object_names:
                continue
            # The target object is the one with less mass.
            if self.object_info[self.object_names[collider_id]].mass < \
                    self.object_info[self.object_names[collidee_id]].mass:
                target = collider_id
                other = collidee_id
            else:
                target = collidee_id
                other = collider_id

            target_audio = self.object_info[self.object_names[target]]
            other_audio = self.object_info[self.object_names[other]]
            commands.append(self.get_impact_sound_command(collision=collision,
                                                          rigidbodies=rigidbodies,
                                                          target_id=target,
                                                          target_amp=target_audio.amp,
                                                          target_mat=target_audio.material.name + "_" + str(target_audio.size),
                                                          other_id=other,
                                                          other_amp=other_audio.amp,
                                                          other_mat=other_audio.material.name + "_" + str(other_audio.size),
                                                          resonance=target_audio.resonance,
                                                          play_audio_data=not resonance_audio))
        # Play sounds from collisions with the environment.
        stays_and_exits.extend([c.get_object_id() for c in env_collisions if
                                c.get_state() == "stay" or c.get_state() == "exit"])
        for collision in env_collisions:
            target = collision.get_object_id()
            # Ignore collisions that aren't enter, not a floor, that aren't in the cached data, or that are too slow.
            # When objects are initially spawned they collide with the environment at very slow speeds,
            # resulting in a "click" sound that we don't actually want.
            if collision.get_state() != "enter" or not collision.get_floor() or target not in self.object_names or \
                    speeds[target] < 0.01 or target in colliders or target in stays_and_exits:
                continue
            colliders.append(target)
            audio = self.object_info[self.object_names[target]]
            commands.append(self.get_impact_sound_command(collision=collision,
                                                          rigidbodies=rigidbodies,
                                                          target_id=target,
                                                          target_amp=audio.amp,
                                                          target_mat=audio.material.name + "_" + str(audio.size),
                                                          other_id=self.env_id,
                                                          other_amp=0.5,
                                                          # We probably need dedicated wall and floor materials, or maybe they are in size category #6?
                                                          # Setting to "4" for now, for general debugging purposes
                                                          other_mat=floor.name + "_4",
                                                          resonance=audio.resonance,
                                                          play_audio_data=not resonance_audio))
        return commands

    def get_log(self) -> dict:
        """
        :return: The mode properties log.
        """

        return self.mode_properties_log

    def _get_object_modes(self, material: Union[str, AudioMaterial]) -> Modes:
        """
        :param material: The audio material.

        :return: The audio modes.
        """
        data = self.material_data[material] if isinstance(material, str) else self.material_data[material.name]
        # Load the mode properties.
        f = -1
        p = -1
        t = -1
        for jm in range(0, 10):
            jf = 0
            while jf < 20:
                jf = data["cf"][jm] + np.random.normal(0, data["cf"][jm] / 10)
            jp = data["op"][jm] + np.random.normal(0, 10)
            jt = 0
            while jt < 0.001:
                jt = data["rt"][jm] + np.random.normal(0, data["rt"][jm] / 10)
            if jm == 0:
                f = jf
                p = jp
                t = jt * 1e3
            else:
                f = np.append(f, jf)
                p = np.append(p, jp)
                t = np.append(t, jt * 1e3)
        return Modes(f, p, t)

    def get_sound(self, collision: Union[Collision, EnvironmentCollision], rigidbodies: Rigidbodies, id1: int, mat1: str, id2: int, mat2: str, other_amp: float, target_amp: float, resonance: float) -> Optional[Base64Sound]:
        """
        Produce sound of two colliding objects as a byte array.

        :param collision: TDW `Collision` or `EnvironmentCollision` output data.
        :param rigidbodies: TDW `Rigidbodies` output data.
        :param id1: The object ID for one of the colliding objects.
        :param mat1: The material label for one of the colliding objects.
        :param id2: The object ID for the other object.
        :param mat2: The material label for the other object.
        :param other_amp: Sound amplitude of object 2.
        :param target_amp: Sound amplitude of object 1.
        :param resonance: The resonances of the objects.

        :return Sound data as a Base64Sound object.
        """

        # The sound amplitude of object 2 relative to that of object 1.
        amp2re1 = other_amp / target_amp

        # Set the object modes.
        if id2 not in self.object_modes:
            self.object_modes.update({id2: {}})
        if id1 not in self.object_modes[id2]:
            self.object_modes[id2].update({id1: CollisionInfo(self._get_object_modes(mat2),
                                                              self._get_object_modes(mat1),
                                                              amp=target_amp * self.initial_amp)})
        obj_col = isinstance(collision, Collision)

        # Unpack useful parameters.
        # Compute normal velocity at impact.
        vel = 0
        if obj_col:
            vel = collision.get_relative_velocity()
        else:
            for i in range(rigidbodies.get_num()):
                if rigidbodies.get_id(i) == id2:
                    vel = rigidbodies.get_velocity(i)
                    # If the y coordinate of the velocity is negative, it implies a scrape or roll along the floor.
                    if vel[1] < 0:
                        return None
                    break
        vel = np.asarray(vel)
        speed = np.square(vel)
        speed = np.sum(speed)
        speed = math.sqrt(speed)
        nvel = vel / np.linalg.norm(vel)
        num_contacts = collision.get_num_contacts()
        nspd = []
        for jc in range(0, num_contacts):
            tmp = np.asarray(collision.get_contact_normal(jc))
            tmp = tmp / np.linalg.norm(tmp)
            tmp = np.arccos(np.clip(np.dot(tmp, nvel), -1.0, 1.0))
            # Scale the speed by the angle (i.e. we want speed Normal to the surface).
            tmp = speed * np.cos(tmp)
            nspd.append(tmp)
        normal_speed = np.mean(nspd)
        # Get indices of objects in collisions.
        id1_index = None
        id2_index = None

        for i in range(rigidbodies.get_num()):
            if rigidbodies.get_id(i) == id1:
                id1_index = i
            if rigidbodies.get_id(i) == id2:
                id2_index = i

        # Use default values for environment collisions.
        if not obj_col:
            m1 = 100
            m2 = rigidbodies.get_mass(id2_index)
        # Use the Rigidbody masses.
        elif id1_index is not None and id2_index is not None:
            m1 = rigidbodies.get_mass(id1_index)
            m2 = rigidbodies.get_mass(id2_index)
        # Fallback: Try to use default mass values if the ID's aren't in the Rigidbody data.
        elif id1 in self.object_names and id2 in self.object_names and self.object_names[id1] in self.object_info and \
                self.object_names[id2] in self.object_info:
            m1 = self.object_info[self.object_names[id1]].mass
            m2 = self.object_info[self.object_names[id2]].mass
        # Failed to generate a sound.
        else:
            return None
        mass = np.min([m1, m2])

        # Re-scale the amplitude.
        if self.object_modes[id2][id1].count == 0:
            # Sample the modes.
            sound, modes_1, modes_2 = self.make_impact_audio(amp2re1, mass, mat1=mat1, mat2=mat2, id1=id1, id2=id2, resonance=resonance)
            # Save collision info - we will need for later collisions.
            amp = self.object_modes[id2][id1].amp
            self.object_modes[id2][id1].init_speed = normal_speed
            self.object_modes[id2][id1].obj1_modes = modes_1
            self.object_modes[id2][id1].obj2_modes = modes_2

        else:
            amp = self.object_modes[id2][id1].amp * normal_speed / self.object_modes[id2][id1].init_speed
            # Adjust modes here so that two successive impacts are not identical.
            modes_1 = self.object_modes[id2][id1].obj1_modes
            modes_2 = self.object_modes[id2][id1].obj2_modes
            modes_1.powers = modes_1.powers + np.random.normal(0, 2, len(modes_1.powers))
            modes_2.powers = modes_2.powers + np.random.normal(0, 2, len(modes_2.powers))
            sound = PyImpact.synth_impact_modes(modes_1, modes_2, mass, resonance)
            self.object_modes[id2][id1].obj1_modes = modes_1
            self.object_modes[id2][id1].obj2_modes = modes_2

        if self.logging:
            mode_props = dict()
            self.log_modes(self.object_modes[id2][id1].count, mode_props, id1, id2, modes_1, modes_2, amp, str(mat1), str(mat2))
            
        # On rare occasions, it is possible for PyImpact to fail to generate a sound.
        if sound is None:
            return None

        # Count the collisions.
        self.object_modes[id2][id1].count_collisions()

        # Prevent distortion by clamping the amp.
        if self.prevent_distortion and np.abs(amp) > 0.99:
            amp = 0.99

        sound = amp * sound / np.max(np.abs(sound))
        return Base64Sound(sound)

    def get_impact_sound_command(self, collision: Union[Collision, EnvironmentCollision], rigidbodies: Rigidbodies, target_id: int, target_mat: str, target_amp: float, other_id: int, other_mat: str, other_amp: float, resonance: float, play_audio_data: bool = True) -> dict:
        """
        Create an impact sound, and return a valid command to play audio data in TDW.
        "target" should usually be the smaller object, which will play the sound.
        "other" should be the larger (stationary) object.

        :param collision: TDW `Collision` or `EnvironmentCollision` output data.
        :param target_amp: The target's amp value.
        :param target_mat: The target's audio material.
        :param other_amp: The other object's amp value.
        :param other_id: The other object's ID.
        :param other_mat: The other object's audio material.
        :param rigidbodies: TDW `Rigidbodies` output data.
        :param target_id: The ID of the object that will play the sound.
        :param resonance: The resonance of the objects.
        :param play_audio_data: If True, return a `play_audio_data` command. If False, return a `play_point_source_data` command (useful only with Resonance Audio; see Command API).

        :return A `play_audio_data` or `play_point_source_data` command that can be sent to the build via `Controller.communicate()`.
        """

        impact_audio = self.get_sound(collision, rigidbodies, other_id, other_mat, target_id, target_mat, other_amp, target_amp, resonance)
        if impact_audio is not None:
            return {"$type": "play_audio_data" if play_audio_data else "play_point_source_data",
                    "id": target_id,
                    "num_frames": impact_audio.length,
                    "num_channels": 1,
                    "frame_rate": 44100,
                    "wav_data": impact_audio.wav_str,
                    "y_pos_offset": 0.1}
        # If PyImpact failed to generate a sound (which is rare!), fail silently here.
        else:
            return {"$type": "do_nothing"}

    def make_impact_audio(self, amp2re1: float, mass: float, id1: int, id2: int, resonance: float, mat1: str = 'cardboard', mat2: str = 'cardboard') -> (np.array, Modes, Modes):
        """
        Generate an impact sound.

        :param mat1: The material label for one of the colliding objects.
        :param mat2: The material label for the other object.
        :param amp2re1: The sound amplitude of object 2 relative to that of object 1.
        :param mass: The mass of the smaller of the two colliding objects.
        :param id1: The ID for the one of the colliding objects.
        :param id2: The ID for the other object.
        :param resonance: The resonance of the objects.

        :return The sound, and the object modes.
        """

        # Unpack material names.
        for jmat in range(0, len(AudioMaterial)):
            if mat1 == AudioMaterial(jmat):
                tmp1 = AudioMaterial(jmat)
                mat1 = tmp1.name
            if mat2 == AudioMaterial(jmat):
                tmp2 = AudioMaterial(jmat)
                mat2 = tmp2.name
        # Sample modes of object1.
        modes_1 = self.object_modes[id2][id1].obj1_modes
        modes_2 = self.object_modes[id2][id1].obj2_modes
        # Scale the two sounds as specified.
        modes_2.decay_times = modes_2.decay_times + 20 * np.log10(amp2re1)
        snth = PyImpact.synth_impact_modes(modes_1, modes_2, mass, resonance)
        return snth, modes_1, modes_2

    def get_impulse_response(self, collision: Union[Collision, EnvironmentCollision], rigidbodies: Rigidbodies, other_id: int, other_mat: str, target_id: int, target_mat: str, other_amp: float, target_amp: float, resonance: float) -> np.array:
        """
        Generate an impulse response from the modes for two specified objects.

        :param collision: TDW `Collision` or `EnvironmentCollision` output data.
        :param target_mat: The target's audio material.
        :param other_id: The other object's ID.
        :param other_mat: The other object's audio material.
        :param rigidbodies: TDW `Rigidbodies` output data.
        :param target_id: The ID of the object that will play the sound.
        :param other_amp: Sound amplitude of other object.
        :param target_amp: Sound amplitude of target object.
        :param resonance: The resonance of the objects.

        :return The impulse response.
        """
        self.get_sound(collision, rigidbodies, other_id, other_mat, target_id, target_mat, other_amp, target_amp, resonance)

        modes_1 = self.object_modes[target_id][other_id].obj1_modes
        modes_2 = self.object_modes[target_id][other_id].obj2_modes
        h1 = modes_1.sum_modes(resonance=resonance)
        h2 = modes_2.sum_modes(resonance=resonance)
        h = Modes.mode_add(h1, h2)
        return h, min(modes_1.frequencies)

    @staticmethod
    def synth_impact_modes(modes1: Modes, modes2: Modes, mass: float, resonance: float) -> np.array:
        """
        Generate an impact sound from specified modes for two objects, and the mass of the smaller object.

        :param modes1: Modes of object 1. A numpy array with: column1=mode frequencies (Hz); column2=mode onset powers in dB; column3=mode RT60s in milliseconds;
        :param modes2: Modes of object 2. Formatted as modes1/modes2.
        :param mass: the mass of the smaller of the two colliding objects.
        :param resonance: The resonance of the objects.

        :return The impact sound.
        """

        h1 = modes1.sum_modes(resonance=resonance)
        h2 = modes2.sum_modes(resonance=resonance)
        h = Modes.mode_add(h1, h2)
        if len(h) == 0:
            return None
        # Convolve with force, with contact time scaled by the object mass.
        max_t = 0.001 * mass
        # A contact time over 2ms is unphysically long.
        max_t = np.min([max_t, 2e-3])
        n_pts = int(np.ceil(max_t * 44100))
        tt = np.linspace(0, np.pi, n_pts)
        frc = np.sin(tt)
        x = sg.fftconvolve(h, frc)
        x = x / abs(np.max(x))
        return x

    @staticmethod
    def get_object_info(csv_file: Union[str, Path] = "") -> Dict[str, ObjectInfo]:
        """
        Returns ObjectInfo values.
        As of right now, only a few objects in the TDW model libraries are included. More will be added in time.

        :param csv_file: The path to the .csv file containing the object info. By default, it will load `tdw/py_impact/objects.csv`. If you want to make your own spreadsheet, use this file as a reference.

        :return: A list of default ObjectInfo. Key = the name of the model. Value = object info.
        """

        objects: Dict[str, ObjectInfo] = {}
        # Load the objects.csv metadata file.
        if isinstance(csv_file, str):
            # Load the default file.
            if csv_file == "":
                csv_file = str(Path(resource_filename(__name__, f"py_impact/objects.csv")).resolve())
            else:
                csv_file = str(Path(csv_file).resolve())
        else:
            csv_file = str(csv_file.resolve())

        # Parse the .csv file.
        with io.open(csv_file, newline='', encoding='utf-8-sig') as f:
            reader = DictReader(f)
            for row in reader:
                o = ObjectInfo(name=row["name"], amp=float(row["amp"]), mass=float(row["mass"]),
                               material=AudioMaterial[row["material"]], library=row["library"],
                               bounciness=float(row["bounciness"]), resonance=float(row["resonance"]), size=int(row["size"]))
                objects.update({o.name: o})

        return objects

    @staticmethod
    def get_collisions(resp: List[bytes]) -> Tuple[List[Collision], List[EnvironmentCollision], Optional[Rigidbodies]]:
        """
        Parse collision and rigibody data from the output data.

        :param resp: The response from the build.

        :return: A list of collisions on this frame (can be empty), a list of environment collisions on this frame (can be empty), and Rigidbodies data (can be `None`).
        """

        if len(resp) == 1:
            return [], [], None
        collisions: List[Collision] = []
        environment_collisions: List[EnvironmentCollision] = []
        rigidbodies: Optional[Rigidbodies] = None
        for r in resp[:-1]:
            r_id = OutputData.get_data_type_id(r)
            if r_id == 'coll':
                collisions.append(Collision(r))
            if r_id == 'rigi':
                rigidbodies = Rigidbodies(r)
            if r_id == 'enco':
                environment_collisions.append(EnvironmentCollision(r))

        return collisions, environment_collisions, rigidbodies

    @staticmethod
    def is_valid_collision(collision: Union[Optional[Collision], Optional[EnvironmentCollision]]) -> bool:
        """
        :param collision: Collision or EnvironmentCollision output data from the build.

        :return: True if this collision can be used to generate an impact sound.
        """

        return collision is not None and ((isinstance(collision, Collision) and
                                           np.linalg.norm(collision.get_relative_velocity()) > 0) or
                                          isinstance(collision, EnvironmentCollision))

    def reset(self, initial_amp: float = 0.5) -> None:
        """
        Reset PyImpact. This is somewhat faster than creating a new PyImpact object per trial.

        :param initial_amp: The initial amplitude, i.e. the "master volume". Must be > 0 and < 1.
        """

        assert 0 < initial_amp < 1, f"initial_amp is {initial_amp} (must be > 0 and < 1)."

        # Clear the object data.
        self.object_modes.clear()

    def log_modes(self, count: int, mode_props: dict, id1: int, id2: int, modes_1: Modes, modes_2: Modes, amp: float, mat1: str, mat2: str):
        """
        Log mode properties info for a single collision event.

        :param count: Mode count for this material-material collision.
        :param mode_props: Dictionary to log to.
        :param id1: ID of the "other" object.
        :param id2: ID of the "target" object.
        :param modes_1: Modes of the "other" object.
        :param modes_2: Modes of the "target" object.
        :param amp: Adjusted amplitude value of collision.
        :param mat1: Material of the "other" object.
        :param mat2: Material of the "target" object.
        """

        mode_props["modes_count"] = count
        mode_props["other_id"] = id1
        mode_props["target_id"] = id2
        mode_props["amp"] = amp
        mode_props["other_material"] = mat1
        mode_props["target_material"] = mat2
        mode_props["modes_1.frequencies"] = modes_1.frequencies.tolist()
        mode_props["modes_1.powers"] = modes_1.powers.tolist()
        mode_props["modes_1.decay_times"] = modes_1.decay_times.tolist()
        mode_props["modes_2.frequencies"] = modes_2.frequencies.tolist()
        mode_props["modes_2.powers"] = modes_2.powers.tolist()
        mode_props["modes_2.decay_times"] = modes_2.decay_times.tolist()
        self.mode_properties_log[str(id1) + "_" + str(id2) + "__" + str(count)] = mode_props



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class QuitSignal(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsQuitSignal(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = QuitSignal()
        x.Init(buf, n + offset)
        return x

    # QuitSignal
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # QuitSignal
    def Ok(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return bool(self._tab.Get(tdw.flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def QuitSignalStart(builder): builder.StartObject(1)
def QuitSignalAddOk(builder, ok): builder.PrependBoolSlot(0, ok, 0)
def QuitSignalEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class TriggerCollision(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsTriggerCollision(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = TriggerCollision()
        x.Init(buf, n + offset)
        return x

    # TriggerCollision
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # TriggerCollision
    def ColliderId(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # TriggerCollision
    def CollideeId(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # TriggerCollision
    def TriggerId(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # TriggerCollision
    def State(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 1

def TriggerCollisionStart(builder): builder.StartObject(4)
def TriggerCollisionAddColliderId(builder, colliderId): builder.PrependInt32Slot(0, colliderId, 0)
def TriggerCollisionAddCollideeId(builder, collideeId): builder.PrependInt32Slot(1, collideeId, 0)
def TriggerCollisionAddTriggerId(builder, triggerId): builder.PrependInt32Slot(2, triggerId, 0)
def TriggerCollisionAddState(builder, state): builder.PrependUint8Slot(3, state, 1)
def TriggerCollisionEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class ScreenPosition(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsScreenPosition(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = ScreenPosition()
        x.Init(buf, n + offset)
        return x

    # ScreenPosition
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # ScreenPosition
    def AvatarId(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ScreenPosition
    def SensorName(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # ScreenPosition
    def Id(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # ScreenPosition
    def World(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # ScreenPosition
    def Screen(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = o + self._tab.Pos
            from .Vector3 import Vector3
            obj = Vector3()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

def ScreenPositionStart(builder): builder.StartObject(5)
def ScreenPositionAddAvatarId(builder, avatarId): builder.PrependUOffsetTRelativeSlot(0, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(avatarId), 0)
def ScreenPositionAddSensorName(builder, sensorName): builder.PrependUOffsetTRelativeSlot(1, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(sensorName), 0)
def ScreenPositionAddId(builder, id): builder.PrependInt32Slot(2, id, 0)
def ScreenPositionAddWorld(builder, world): builder.PrependStructSlot(3, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(world), 0)
def ScreenPositionAddScreen(builder, screen): builder.PrependStructSlot(4, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(screen), 0)
def ScreenPositionEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class BoundsData(object):
    __slots__ = ['_tab']

    # BoundsData
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # BoundsData
    def Id(self): return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, self._tab.Pos + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # BoundsData
    def Front(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 4)
        return obj

    # BoundsData
    def Back(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 16)
        return obj

    # BoundsData
    def Right(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 28)
        return obj

    # BoundsData
    def Left(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 40)
        return obj

    # BoundsData
    def Top(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 52)
        return obj

    # BoundsData
    def Bottom(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 64)
        return obj

    # BoundsData
    def Center(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 76)
        return obj


def CreateBoundsData(builder, id, front_x, front_y, front_z, back_x, back_y, back_z, right_x, right_y, right_z, left_x, left_y, left_z, top_x, top_y, top_z, bottom_x, bottom_y, bottom_z, center_x, center_y, center_z):
    builder.Prep(4, 88)
    builder.Prep(4, 12)
    builder.PrependFloat32(center_z)
    builder.PrependFloat32(center_y)
    builder.PrependFloat32(center_x)
    builder.Prep(4, 12)
    builder.PrependFloat32(bottom_z)
    builder.PrependFloat32(bottom_y)
    builder.PrependFloat32(bottom_x)
    builder.Prep(4, 12)
    builder.PrependFloat32(top_z)
    builder.PrependFloat32(top_y)
    builder.PrependFloat32(top_x)
    builder.Prep(4, 12)
    builder.PrependFloat32(left_z)
    builder.PrependFloat32(left_y)
    builder.PrependFloat32(left_x)
    builder.Prep(4, 12)
    builder.PrependFloat32(right_z)
    builder.PrependFloat32(right_y)
    builder.PrependFloat32(right_x)
    builder.Prep(4, 12)
    builder.PrependFloat32(back_z)
    builder.PrependFloat32(back_y)
    builder.PrependFloat32(back_x)
    builder.Prep(4, 12)
    builder.PrependFloat32(front_z)
    builder.PrependFloat32(front_y)
    builder.PrependFloat32(front_x)
    builder.PrependInt32(id)
    return builder.Offset()



from typing import List, Union
from tdw.controller import Controller
from tdw.output_data import OutputData, Keyboard


class KeyboardController(Controller):
    """
    Listen for keyboard input to send commands.

    Keyboard input is registered _from the build, not the controller._ For this controller to work, you must:

    - Run the build on the same machine as the keyboard.
    - Have the build window as the focused window (i.e. not minimized).

    Usage:

    ```python
    from tdw.keyboard_controller import KeyboardController
    from tdw.tdw_utils import TDWUtils

    def stop():
        done = True
        c.communicate({"$type": "terminate"})

    done = False
    c = KeyboardController()
    c.start()

    # Quit.
    c.listen(key="esc", function=stop)

    # Equivalent to c.start()
    c.listen(key="r", commands={"$type": "load_scene", "scene_name": "ProcGenScene"})

    while not done:
        # Receive data. Load the scene when r is pressed. Quit when Esc is pressed.
        c.communicate([])
    ```
    """

    def __init__(self, port: int = 1071, check_version: bool = True, launch_build: bool = True):
        """
        Create the network socket and bind the socket to the port.

        :param port: The port number.
        :param check_version: If true, the controller will check the version of the build and print the result.
        :param launch_build: If True, automatically launch the build. If one doesn't exist, download and extract the correct version. Set this to False to use your own build, or (if you are a backend developer) to use Unity Editor.
        """

        # Commands that should be added due to key presses on this frame.
        self.on_key_commands: List[dict] = []

        # Dictionaries of actions. Key = a keyboard key as a string.
        self._press = dict()
        self._hold = dict()
        self._release = dict()

        super().__init__(port=port, check_version=check_version, launch_build=launch_build)
        self.communicate({"$type": "send_keyboard",
                          "frequency": "always"})

    def _do_event(self, k: str, events: dict) -> None:
        """
        Invoke an event if the key is in the events dictionary.

        :param k: The keyboard key as a string.
        :param events: The events dictionary.
        """

        if k in events:
            # Append commands to the next `communicate()` call.
            if isinstance(events[k], list):
                self.on_key_commands.extend(events[k])
            # Invoke a function.
            else:
                events[k]()

    def communicate(self, commands: Union[dict, List[dict]]) -> List[bytes]:
        if isinstance(commands, dict):
            commands = [commands]
        # Add commands from key presses.
        commands.extend(self.on_key_commands[:])
        # Clear the on-key commands.
        self.on_key_commands.clear()
        # Send the commands.
        resp = super().communicate(commands)

        # Get keyboard input.
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "keyb":
                keys = Keyboard(resp[i])
                # Listen for events where the key was first pressed on the previous frame.
                for j in range(keys.get_num_pressed()):
                    self._do_event(k=keys.get_pressed(j), events=self._press)
                # Listen for keys currently held down.
                for j in range(keys.get_num_held()):
                    self._do_event(k=keys.get_held(j), events=self._hold)
                # Listen for keys that were released.
                for j in range(keys.get_num_released()):
                    self._do_event(k=keys.get_released(j), events=self._release)
        return resp

    def listen(self, key: str, commands: Union[dict, List[dict]] = None, function=None, events: List[str] = None) -> None:
        """
        Listen for when a key is pressed and send commands.

        :param key: The keyboard key.
        :param commands: Commands to be sent when the key is pressed.
        :param function: Function to invoke when the key is pressed.
        :param events: Listen to these keyboard events for this `key`. Options: `"press"`, `"hold"`, `"release"`. If None, this defaults to `["press"]`.
        """

        response = None
        if commands is not None:
            if isinstance(commands, dict):
                commands = [commands]
            response = commands
        elif function is not None:
            response = function
        if response is None:
            return
        if events is None:
            events = ["press"]

        # Subscribe to events.
        if "press" in events:
            self._press[key] = response
        if "hold" in events:
            self._hold[key] = response
        if "release" in events:
            self._release[key] = response

    def _set_frame_commands(self, commands: Union[dict, List[dict]]) -> None:
        """
        Set the next frame's commands.

        :param commands: The commands to send on this frame.
        """

        if isinstance(commands, dict):
            commands = [commands]
        self.on_key_commands = commands



from typing import List, Dict, Optional
from os import getcwd, chdir, walk, devnull
import re
from platform import system
from pathlib import Path
from subprocess import call
from distutils.file_util import copy_file, move_file
from distutils.dir_util import remove_tree
from tdw.asset_bundle_creator_base import AssetBundleCreatorBase
from tdw.librarian import RobotRecord
from tdw.backend.paths import EDITOR_LOG_PATH
from tdw.backend.platforms import UNITY_TO_SYSTEM


class RobotCreator(AssetBundleCreatorBase):
    """
    Download a .urdf or .xacro file and convert it into an asset bundle that is usable by TDW.

    # Requirements

    - Windows 10, OS X, or Linux
      - On a remote Linux server, you'll need a valid virtual display (see the `display` parameter of the constructor)
    - Unity Editor 2020.2 (must be installed via Unity Hub)
    - Python3 and the `tdw` module
    - git

    ### ROS and .xacro file requirements

    If you want to use a .xacro file, `RobotCreator` can convert it to a usable .urdf file, provided that you first install ROS. If you already have a .urdf file, you don't need to install ROS.

    [This is the source of the installation instructions listed below.](http://wiki.ros.org/Installation/Ubuntu)

    On Linux:

    - `sudo apt-key adv --keyserver 'hkp://keyserver.ubuntu.com:80' --recv-key C1CF6E31E6BADE8868B172B4F42ED6FBAB17C654`
    - `curl -sSL 'http://keyserver.ubuntu.com/pks/lookup?op=get&search=0xC1CF6E31E6BADE8868B172B4F42ED6FBAB17C654' | sudo apt-key add -`
    - `sudo ros-melodic-ros-base`
    - `sudo apt-get install rosbash`
    - `sudo apt-get install ros-melodic-xacro`

    On Windows:

    - Install Ubuntu 18 on WSL 2
    - `wsl sudo apt-key adv --keyserver 'hkp://keyserver.ubuntu.com:80' --recv-key C1CF6E31E6BADE8868B172B4F42ED6FBAB17C654`
    - `wsl curl -sSL 'http://keyserver.ubuntu.com/pks/lookup?op=get&search=0xC1CF6E31E6BADE8868B172B4F42ED6FBAB17C654' | sudo apt-key add -`
    - `wsl sudo ros-melodic-ros-base`
    - `wsl sudo apt-get install rosbash`
    - `wsl sudo apt-get install ros-melodic-xacro`

    On OS X:

    ROS isn't well-supported on OS X. You can try following installation instructions [here](http://wiki.ros.org/Installation/).

    # Usage

    To create an asset bundle of the UR5 robot:

    ```python
    from tdw.robot_creator import RobotCreator

    r = RobotCreator()
    record = r.create_asset_bundles(
        urdf_url="https://github.com/ros-industrial/robot_movement_interface/blob/master/dependencies/ur_description/urdf/ur5_robot.urdf",
        xacro_args=None,
        required_repo_urls=None,
        immovable=True,
        up="y")
    print(record.name)
    print(record.urls)
    ```

    Note that most of the parameters are optional, so this can be simplified to:

    ```python
    from tdw.robot_creator import RobotCreator

    r = RobotCreator()
    record = r.create_asset_bundles(urdf_url="https://github.com/ros-industrial/robot_movement_interface/blob/master/dependencies/ur_description/urdf/ur5_robot.urdf")
    print(record.name)
    print(record.urls)
    ```

    The first time that this script is run, it clone [the robot_creator repo](https://github.com/alters-mit/robot_creator) (a Unity project used for creating robots) to your home directory.

    ### Edit the prefab

    `RobotCreator.create_asset_bundles()` creates a .prefab file in the `robot_creator` Unity project. This is an intermediate file that is required for building the asset bundle.

    It's usually worth testing and editing the prefab before finalizing the asset bundle. To do this, first run `RobotCreator` as described above, and then do the following:

    1. Create a prefab of the robot.
    2. Open robot_creator Unity project in Unity 2020.2; the project is located at `~/robot_creator` (where `~` is your home directory).
    3. In the Unity Editor project window, double-click `Scenes -> SampleScene`
    4. In the Unity Editor project window, search for the name of the robot. Click the file and drag it into the scene view.
    5. Press play.

    Common problems and solutions during prefab creation:

    | Problem | Solution |
    | --- | --- |
    | Prefab creation seems to hang. | This is because sometimes the physics hull collider meshes are very complicated and require more time to generate. Let the process run. |
    | Got an error during prefab creation: `Root object of the robot doesn't have an ArticulationBody.` | Open the project, double-click the prefab, and add an ArticulationBody to the root object. Adjust the parenting hierarchy of the robot such that the ArticulationBodies beneath the root are direct children. This is bad: `root -> non-articulation -> articulation` and this is good: `root -> articulation` |

    Common problems and solutions while testing a prefab in the Unity Editor project:

    | Problem | Solution |
    | --- | --- |
    | Arms are flailing. | Usually this is because the colliders are parented to the wrong object. Double-click the prefab and make sure each `Collisions` object is parented to the matching ArticulationBody object. |
    | Robot falls apart and there are `AABB` errors | You have too many ArticulationBodies. Unity supports a maximum of 65 (1 parent, 64 children). Double-click the prefab and delete any redundant ArticulationBodies. |
    | The base of the robot is below (0, 0, 0) | Double-click the prefab and adjust the y position of the child objects. |
    | Joints snap to a weird angle. | Usually this is because there are overlapping physics colliders. Double-click the prefab and in the Hierarchy panel click the root object. The green wireframe meshes in the Scene View are the physics colliders. Try deleting or disabling colliders near the glitching joint. |
    | The robot tips over. | Set `immovable=True` in `create_asset_bundles()`. If that doesn't work, double-click the prefab. In the Hierarchy panel, click the ArticulationBody that you think is causing the robot to tilt. In the Inspector panel, click "Add Component". Add: `Center Of Mass`. Adjust the center of mass in the Inspector until the robot stops tipping. |

    ### Create an asset bundle from a prefab

    Once the robot prefab is working correctly, you can create asset bundles without overwriting the prefab:

    ```python
    from tdw.robot_creator import RobotCreator

    r = RobotCreator()
    urdf_url = "https://github.com/ros-industrial/robot_movement_interface/blob/master/dependencies/ur_description/urdf/ur5_robot.urdf"
    record = r.create_asset_bundles(urdf_url=urdf_url)
    print(record.name)
    print(record.urls)

    asset_bundles = rc.prefab_to_asset_bundles(name=name)
    record = RobotCreator.get_record(name=record.name, urdf_url=urdf_url, immovable=True, asset_bundles=asset_bundles)
    ```

    ### Store metadata

    `RobotCreator.create_asset_bundles()` returns a `RobotRecord` metadata object, which contains the files paths to the asset bundle.
    You can store a `RobotRecord` object in a `RobotLibrarian`, which is saved as a JSON file.

    To create a `RobotLibrarian`:

    ```python
    from tdw.librarian import RobotLibrarian

    RobotLibrarian.create_library(path="my_robot_librarian.json", description="Custom Robot Librarian")
    ```

    To create asset bundles and save the `RobotRecord` metadata:

    ```python
    from tdw.robot_creator import RobotCreator
    from tdw.librarian import RobotLibrarian

    r = RobotCreator()

    # Create the asset bundles and generate a metadata record.
    record = r.create_asset_bundles(urdf_url="https://github.com/ros-industrial/robot_movement_interface/blob/master/dependencies/ur_description/urdf/ur5_robot.urdf")

    # Add the record to the local library.
    lib = RobotLibrarian("my_robot_librarian.json")
    lib.add_or_update_record(record=record, overwrite=False, write=True)
    ```

    ### Load the robot into TDW

    To add the robot to a TDW scene:

    ```python
    from pathlib import Path
    from tdw.controller import Controller
    from tdw.tdw_utils import TDWUtils
    from tdw.librarian import RobotLibrarian
    from tdw.robot_creator import RobotCreator

    lib_path = "my_robot_librarian.json"
    # Create your robot library if it doesn't already exist.
    if not Path(lib_path).exists():
        RobotLibrarian.create_library(path=lib_path, description="Custom Robot Librarian")
    # Load your robot library.
    lib = RobotLibrarian(lib_path)

    robot_name = "ur5"
    # If there isn't a metadata record yet for this robot, create asset bundles and add the record.
    if lib.get_record(robot_name) is None:
        r = RobotCreator()
        record = r.create_asset_bundles(urdf_url="https://github.com/ros-industrial/robot_movement_interface/blob/master/dependencies/ur_description/urdf/ur5_robot.urdf")
        # Add the record if it doesn't already exist.
        lib.add_or_update_record(record=record, overwrite=False, write=True)

    # Launch the controller.
    c = Controller(launch_build=False)
    c.start()
    robot_id = 0
    commands = [TDWUtils.create_empty_room(12, 12),
                c.get_add_robot(name=robot_name, robot_id=robot_id, library=lib_path)]
    commands.extend(TDWUtils.create_avatar(look_at=TDWUtils.VECTOR3_ZERO, position={"x": -0.881, "y": 0.836, "z": -1.396}))
    c.communicate(commands)
    ```

    For a more complete example of how to use the robot in TDW (i.e. find all of its joints and set target angles or positions), see: `twd/Python/example_controllers/robot_arm.py` Note that in this example, there isn't any code to build the asset bundles or update the metadata. This is because UR5 is already in the default TDW robot library (on our remote S3 server) and the metadata record is already stored in the default metadata librarian.

    ***

    # Functions

    """

    # The root temporary directory.
    TEMP_ROOT = Path.home().joinpath("robot_creator/temp_robots")

    def __init__(self, quiet: bool = False, display: str = ":0"):
        """
        :param quiet: If true, don't print any messages to console.
        :param display: The display to launch Unity Editor on. Ignored if this isn't Linux.
        """

        super().__init__(quiet=quiet, display=display)

    def create_asset_bundles(self, urdf_url: str, required_repo_urls: Dict[str, str] = None, xacro_args: Dict[str, str] = None, immovable: bool = True, up: str = "y") -> RobotRecord:
        """
        Given the URL of a .urdf file or a .xacro file, create asset bundles of the robot.

        This is a wrapper function for:

        1. `clone_repo()`
        2. `copy_files()`
        3. `urdf_to_prefab()`
        4. `prefab_to_asset_bundles()`

        :param urdf_url: The URL of a .urdf or a .xacro file.
        :param required_repo_urls: A dictionary of description folder names and repo URLs outside of the robot's repo that are required to create the robot. This is only required for .xacro files that reference outside repos. For example, the Sawyer robot requires this to add the gripper: `{"intera_tools_description": "https://github.com/RethinkRobotics/intera_common"}`
        :param xacro_args: Names and values for the `arg` tags in the .xacro file (ignored if this is a .urdf file). For example, the Sawyer robot requires this to add the gripper: `{"electric_gripper": "true"}`
        :param immovable: If True, the base of the robot is immovable.
        :param up: The up direction. Used when importing the robot into Unity. Options: `"y"` or `"z"`. Usually, this should be the default value (`"y"`).

        :return: A `RobotRecord` object. The `urls` field contains the paths to each asset bundle.
        """

        if required_repo_urls is None:
            required_repo_urls = list()

        # Clone the repo.
        repo_paths: Dict[str, Path] = dict()
        local_repo_path = self.clone_repo(url=urdf_url)
        repo_paths[RobotCreator._get_description_infix(url=urdf_url)] = local_repo_path

        # Clone the required repos.
        for description in required_repo_urls:
            required_repo_url = required_repo_urls[description]
            required_local_repo_path = self.clone_repo(url=required_repo_url)
            repo_paths[description] = required_local_repo_path

        # Copy the files, create a .urdf file (if needed), and creator collider objects.
        urdf_path = self.copy_files(urdf_url=urdf_url, local_repo_path=local_repo_path, repo_paths=repo_paths,
                                    xacro_args=xacro_args)

        # Create the prefab.
        prefab = self.urdf_to_prefab(urdf_path=urdf_path, immovable=immovable, up=up)
        name = prefab.name.replace(".prefab", "")

        # Create the asset bundles.
        asset_bundles = self.prefab_to_asset_bundles(name=name)
        temp = dict()
        for k in asset_bundles:
            temp[k] = "file:///" + str(asset_bundles[k].resolve()).replace("\\", "/")
        asset_bundles = temp

        # Create the record.
        return RobotCreator.get_record(name=name, urdf_url=urdf_url, immovable=immovable, asset_bundles=asset_bundles)

    @staticmethod
    def get_record(name: str, urdf_url: str, immovable: bool, asset_bundles: Dict[str, str]) -> RobotRecord:
        """
        :param name: The name of the robot.
        :param urdf_url: The URL to the .urdf or .xacro file.
        :param immovable: If True, the base of the robot is immovable.
        :param asset_bundles: The paths to the asset bundles. See `prefab_to_asset_bundles()`.

        :return: A `RobotRecord` metadata object.
        """

        record_data = {"name": name,
                       "source": RobotCreator._get_repo_url(url=urdf_url),
                       "immovable": immovable,
                       "urls": asset_bundles}
        return RobotRecord(data=record_data)

    def clone_repo(self, url: str) -> Path:
        """
        Clone a repo to a temporary directory.

        :param url: The URL to the .urdf or .xacro file or the repo.

        :return: The temporary directory.
        """

        if not RobotCreator.TEMP_ROOT.exists():
            RobotCreator.TEMP_ROOT.mkdir(parents=True)

        # This is a .urdf or .xacro file. Parse the repo URL accordingly.
        if url.endswith(".xacro") or url.endswith(".urdf"):
            local_repo_path = RobotCreator._get_local_repo_path(url=url)
            repo_url = RobotCreator._get_repo_url(url=url)
        # This is the base URL of the repo. Parse it accordingly.
        else:
            local_repo_path = RobotCreator.TEMP_ROOT.joinpath(Path(url).name)
            repo_url = url
        if local_repo_path.exists():
            return local_repo_path

        # Change directory.
        cwd = getcwd()
        chdir(str(RobotCreator.TEMP_ROOT.resolve()))
        if not self.quiet:
            print(f"Cloning: {repo_url}")
        # Clone the repo.
        call(["git", "clone", repo_url],
             stderr=open(devnull, "wb"))
        chdir(cwd)
        assert local_repo_path.exists(), f"Can't find: {local_repo_path.resolve()}"
        if not self.quiet:
            print("...Done!")
        return local_repo_path

    def copy_files(self, urdf_url: str, local_repo_path: Path, repo_paths: Dict[str, Path], xacro_args: Dict[str, str] = None) -> Path:
        """
        Copy and convert files required to create a prefab.

        1. If this is a .xacro file, convert it to a .urdf file.
        2. Copy the .urdf file to the Unity project.
        3. Copy all associated meshes to the .urdf project.

        :param urdf_url: The URL to the remote .urdf or .xacro file.
        :param local_repo_path: The path to the local repo.
        :param repo_paths: A dictionary of required repos (including the one that the .urdf or .xacro is in). Key = The description path infix, e.g. "sawyer_description". Value = The path to the local repo.
        :param xacro_args: Names and values for the `arg` tags in the .xacro file. Can be None for a .urdf or .xacro file and always ignored for a .urdf file.

        :return: The path to the .urdf file in the Unity project.
        """

        page_url = self._raw_to_page(url=urdf_url)
        repo_path = re.search(r"(.*)/blob/master/(.*)", page_url).group(2)
        urdf_path = local_repo_path.joinpath(repo_path)
        dst_root = self.project_path.joinpath(f"Assets/robots")
        if not dst_root.exists():
            dst_root.mkdir(parents=True)
        # Convert the .xacro file to a .urdf file.
        if Path(urdf_url).suffix == ".xacro":
            urdf_dst = self.xacro_to_urdf(xacro_path=urdf_path, repo_paths=repo_paths, args=xacro_args)
        # Move the existing .urdf file.
        else:
            assert urdf_path.exists(), f"Not found: {urdf_path.resolve()}"
            # Copy the .urdf file.
            urdf_dst = dst_root.joinpath(Path(urdf_url).name)
            copy_file(src=str(urdf_path.resolve()), dst=str(urdf_dst.resolve()))

        # Read the .urdf file.
        urdf = urdf_dst.read_text(encoding="utf-8")
        # Remove gazebo stuff.
        urdf = re.sub(r"<gazebo reference(.*?)>((.|\n)*?)</gazebo>", "", urdf)
        urdf = re.sub(r"<gazebo>((.|\n)*?)</gazebo>", "", urdf)
        urdf = re.sub(r"<transmission((.|\n)*?)</transmission>", "", urdf)
        urdf = re.sub(r'<xacro:include (.*)gazebo(.*)/>', "", urdf)
        urdf_dst.write_text(urdf, encoding="utf-8")
        # Copy the meshes.
        for m in re.findall(r"filename=\"package://((.*)\.(DAE|dae|stl|STL))\"", urdf):
            mesh_description = m[0].split("/")[0]
            mesh_repo: Optional[Path] = None
            mesh_desc = ""
            for k_desc in repo_paths:
                if mesh_description in k_desc:
                    mesh_repo: Path = repo_paths[k_desc]
                    mesh_desc = k_desc
                    break
            if mesh_repo is None:
                raise Exception(f"Couldn't find local repo for: {m[0]}")
            if "/" in mesh_desc:
                mesh_src = mesh_repo.joinpath(mesh_desc.replace(mesh_desc.split("/")[-1], "")).joinpath(m[0])
            else:
                mesh_src = mesh_repo.joinpath(m[0])
            assert mesh_src.exists(), f"Not found: {mesh_src}"
            mesh_dst = dst_root.joinpath(m[0])
            if not mesh_dst.parent.exists():
                mesh_dst.parent.mkdir(parents=True)
            # Copy the mesh file to the Unity project.
            copy_file(src=str(mesh_src.resolve()), dst=str(mesh_dst.resolve()))
        if not self.quiet:
            print("Copied the .urdf and the meshes to the Unity project.")
        return urdf_dst

    def xacro_to_urdf(self, xacro_path: Path, repo_paths: Dict[str, Path], args: Dict[str, str] = None) -> Path:
        """
        Convert a local .xacro file to a .urdf file.

        :param xacro_path: The path to the local .xacro file.
        :param args: Names and values for the `arg` tags in the .xacro file.
        :param repo_paths: Local paths to all required repos. Key = The description infix. Value = The local repo path.

        :return: The path to the .urdf file.
        """

        if not RobotCreator.TEMP_ROOT.exists():
            RobotCreator.TEMP_ROOT.mkdir(parents=True)

        xacro = xacro_path.read_text(encoding="utf-8")

        # Set the args.
        if args is None:
            args = {"gazebo": 'false'}
        for k in args:
            xacro = re.sub('<xacro:arg name="' + k + '" default="(.*)"',
                           f'<xacro:arg name="{k}" default="{args[k]}"', xacro)
        xacro = re.sub(r'<xacro:include (.*)gazebo(.*?)/>', "", xacro)

        # Put all required .xacro files in a temporary directory.
        xacro_dir = RobotCreator.TEMP_ROOT.joinpath("xacro")
        if not xacro_dir.exists():
            xacro_dir.mkdir(parents=True)
        x = xacro_dir.joinpath(xacro_path.name)
        x.write_text(xacro, encoding="utf-8")

        xacros: List[Path] = [x]
        checked: List[Path] = []
        while len(xacros) > 0:
            xp = xacros.pop(0)
            checked.append(xp)
            xacro = xp.read_text(encoding="utf-8")
            for description in re.findall(r"\$\(find (.*?)\)", xacro, flags=re.MULTILINE):
                xacro_repo: Optional[Path] = None
                desc = ""
                for k_desc in repo_paths:
                    if description in k_desc:
                        xacro_repo = repo_paths[k_desc]
                        desc = k_desc
                        break
                assert xacro_repo is not None, f"Couldn't find: {description} in {xacro_repo} for {xp}"
                src_urdf_dir = xacro_repo.joinpath(desc).joinpath("urdf")
                for root_dir, dirs, files in walk(str(src_urdf_dir.resolve())):
                    for f in files:
                        src = Path(root_dir).joinpath(f)
                        if src.is_file() and src.suffix == ".xacro":
                            dst = xacro_dir.joinpath(src.name)
                            if not dst.exists():
                                copy_file(src=str(src.resolve()), dst=str(dst.resolve()))
                            if src not in xacros and src not in checked:
                                xacros.append(src)
        if not self.quiet:
            print("Copied all required xacro files to a temp directory.")
        # "Repair" all of the required .xacro files.
        for f in xacro_dir.iterdir():
            if f.is_file() and f.suffix == ".xacro":
                xacro = f.read_text(encoding="utf-8")
                xacro = re.sub(r"include filename=\"\$\((.*)\)/(.*)/(.*)\"", r'include filename="\3"', xacro)
                f.write_text(xacro, encoding="utf-8")
        # Finally, create the .urdf file.
        cwd = getcwd()
        chdir(str(RobotCreator.TEMP_ROOT.joinpath("xacro").resolve()))
        urdf_name = x.name.replace(".urdf.xacro", ".urdf").replace(".xacro", ".urdf")
        xacro_call = ["source", "/opt/ros/melodic/setup.bash", "&&",
                      "rosrun", "xacro", "xacro", "-o", urdf_name, x.name]
        if system() == "Windows":
            xacro_call.insert(0, "wsl")
        call(xacro_call)
        urdf_path = Path(f"../../Assets/robots/{urdf_name}")
        if urdf_path.exists():
            urdf_path.unlink()
        move_file(src=str(x.parent.joinpath(urdf_name).resolve()), dst=str(urdf_path.resolve()))
        if not self.quiet:
            print(f"Created {str(urdf_path.resolve())}")
        urdf_path = Path(str(urdf_path.resolve()))
        chdir(cwd)
        # Delete temp xacro files.
        remove_tree(str(xacro_dir.resolve()))
        assert urdf_path.exists(), f"Not found: {urdf_path.resolve()}"
        return urdf_path.resolve()

    def urdf_to_prefab(self, urdf_path: Path, immovable: bool = True, up: str = "y") -> Path:
        """
        Convert a .urdf file to Unity prefab.

        The .urdf file must already exist on this machine and its meshes must be at the expected locations.

        :param urdf_path: The path to the .urdf file.
        :param immovable: If True, the base of the robot will be immovable by default (see the `set_immovable` command).
        :param up: The up direction. Used for importing the .urdf into Unity. Options: "y" or "z".

        :return: The path to the .prefab file.
        """

        # Get the expected name of the robot.
        urdf = urdf_path.read_text(encoding="utf-8")
        name = re.search(r'<robot name="(.*?)"', urdf, flags=re.MULTILINE).group(1).strip()

        urdf_call = self.get_base_unity_call()[:]
        urdf_call.extend(["-executeMethod", "Creator.CreatePrefab",
                          f"-urdf='{str(urdf_path.resolve())}'",
                          f"-immovable={'true' if immovable else 'false'}",
                          f"-up={up}"])
        if not self.quiet:
            print("Creating a .prefab from a .urdf file...")
        call(urdf_call)
        RobotCreator._check_log()
        prefab_path = self.project_path.joinpath(f"Assets/prefabs/{name}.prefab")
        assert prefab_path.exists(), f"Prefab not found: {prefab_path}"
        if not self.quiet:
            print("...Done!")
        return prefab_path

    def prefab_to_asset_bundles(self, name: str) -> Dict[str, Path]:
        """
        Create asset bundles from a prefab.

        :param name: The name of the robot (minus the .prefab extension).

        :return: A dictionary. Key = The system platform. Value = The path to the asset bundle as a Path object.
        """

        asset_bundles_call = self.get_base_unity_call()[:]
        asset_bundles_call.extend(["-executeMethod", "Creator.CreateAssetBundles",
                                   f"-robot='{name}'"])
        if not self.quiet:
            print("Creating asset bundles...")
        call(asset_bundles_call)
        RobotCreator._check_log()
        # Verify that the asset bundles exist.
        asset_bundles_root_dir = self.project_path.joinpath(f"Assets/asset_bundles/{name}")
        asset_bundle_paths: Dict[str, Path] = dict()
        for build_target in UNITY_TO_SYSTEM:
            asset_bundle_path = asset_bundles_root_dir.joinpath(f"{build_target}/{name}")
            asset_bundle_path = Path(str(Path(asset_bundle_path.resolve())))
            assert asset_bundle_path.exists(), f"Couldn't find asset bundle: {asset_bundle_path.resolve()}"
            asset_bundle_paths[UNITY_TO_SYSTEM[build_target]] = asset_bundle_path
        if not self.quiet:
            print("...Done!")
        return asset_bundle_paths

    def get_unity_project(self) -> Path:
        """
        Build the asset_bundle_creator Unity project.

        :return The path to the asset_bundle_creator Unity project.
        """

        unity_project_path = self.get_project_path()

        # If the project already exists, stop.
        if unity_project_path.exists():
            assert unity_project_path.joinpath("Assets").exists(), f"There is a directory at {unity_project_path} but it's not a Unity project."
            return unity_project_path

        # Clone the repo.
        cwd = getcwd()
        chdir(str(Path.home().resolve()))
        call(["git", "clone", "https://github.com/alters-mit/robot_creator"])
        chdir(cwd)
        assert unity_project_path.exists(), f"Can't find project path: {unity_project_path}"
        return unity_project_path

    @staticmethod
    def get_project_path() -> Path:
        """
        :return: The expected path of the Unity project.
        """

        return Path.home().joinpath("robot_creator")

    @staticmethod
    def _check_log() -> None:
        """
        Check the Editor log for errors.
        """

        log = EDITOR_LOG_PATH.read_text(encoding="utf-8")
        if "failure" in log.lower() or "exception" in log.lower():
            raise Exception(f"There are errors in the Editor log!")

    @staticmethod
    def _page_to_raw(url: str) -> str:
        """
        Convert the URL of a GitHub page to the URL of the corresponding text file.

        :param url: A URL to a GitHub page.

        :return: The URL to the corresponding text file.
        """

        if "https://github.com" in url:
            return re.sub(r"https://github\.com/(.*)/blob/(.*)", r"https://raw.githubusercontent.com/\1/\2", url)
        elif "https://raw.githubusercontent.com" in url:
            return url
        else:
            raise Exception(f"Unexpected URL: {url}")

    @staticmethod
    def _raw_to_page(url: str):
        """
        Convert the URL of a raw text file to the corresponding GitHub page.

        :param url: A URL to a text file page.

        :return: The URL to the corresponding GitHub page.
        """

        if "https://github.com" in url:
            return url
        elif "https://raw.githubusercontent.com" in url:
            return re.sub(r"https://raw\.githubusercontent\.com/(.*)/master/(.*)",
                          r"https://github.com/\1/blob/master/\2", url)
        else:
            raise Exception(f"Unexpected URL: {url}")

    @staticmethod
    def _get_repo_url(url: str) -> str:
        """
        :param url: The URL of the .urdf or .xacro file.

        :return: The base repo of a .urdf or .xacro file.
        """

        page_url = RobotCreator._raw_to_page(url=url)
        return re.sub(r"https://github\.com/(.*)/blob/(.*)", r"https://github.com/\1", page_url)

    @staticmethod
    def _get_repo_name(repo_url: str) -> str:
        """
        :param repo_url: The base URL of a repo.

        :return: The expected name of the repo.
        """

        return repo_url.split("/")[-1]

    @staticmethod
    def _get_local_repo_path(url: str) -> Path:
        """
        :param url: The URL of the .urdf or .xacro file.

        :return: The path to the local repo.
        """

        if not RobotCreator.TEMP_ROOT.exists():
            RobotCreator.TEMP_ROOT.mkdir(parents=True)

        repo_url = RobotCreator._get_repo_url(url=url)
        repo_name = RobotCreator._get_repo_name(repo_url=repo_url)
        return RobotCreator.TEMP_ROOT.joinpath(repo_name)

    @staticmethod
    def _get_description_infix(url: str) -> str:
        """
        :param url: The URL of the .urdf or .xacro file.

        :return: The string between the repo URL and the /urdf/ directory.
        """

        page = RobotCreator._raw_to_page(url=url)
        s = re.search(r"(.*)/blob/master/(.*)/urdf", page)
        if s is None:
            return re.search(r"(.*)/blob/master/((.*)_description)/", page).group(2)
        else:
            return s.group(2)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class MagnebotWheels(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsMagnebotWheels(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = MagnebotWheels()
        x.Init(buf, n + offset)
        return x

    # MagnebotWheels
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # MagnebotWheels
    def Id(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # MagnebotWheels
    def Success(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(self._tab.Get(tdw.flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def MagnebotWheelsStart(builder): builder.StartObject(2)
def MagnebotWheelsAddId(builder, id): builder.PrependInt32Slot(0, id, 0)
def MagnebotWheelsAddSuccess(builder, success): builder.PrependBoolSlot(1, success, 0)
def MagnebotWheelsEnd(builder): return builder.EndObject()



from typing import List, Optional
from tdw.output_data import OutputData, Environments
from tdw.scene.room_bounds import RoomBounds


class SceneBounds:
    """
    Data for the scene bounds and its rooms.

    In order to initialize this object, the controller must have sent `send_environments` to the build on the previous frame:

    ```python
    from tdw.controller import Controller
    from tdw.tdw_utils import TDWUtils
    from tdw.scene.scene_bounds import SceneBounds

    c = Controller()
    c.start()
    resp = c.communicate([TDWUtils.create_empty_room(12, 12),
                          {"$type": "send_environments"}])
    scene_bounds = SceneBounds(resp=resp)
    ```
    """

    def __init__(self, resp: List[bytes]):
        """
        :param resp: The response from the build.
        """

        env: Optional[Environments] = None
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "envi":
                env = Environments(resp[i])
                break
        assert env is not None, "No scene data in response from build!"

        # Get the overall size of the scene.
        """:field
        Minimum x positional coordinate of the scene.
        """
        self.x_min: float = 1000
        """:field
        Maximum x positional coordinate of the scene.
        """
        self.x_max: float = 0
        """:field
        Minimum y positional coordinate of the scene.
        """
        self.y_min: float = 1000
        """:field
        Maximum y positional coordinate of the scene.
        """
        self.y_max: float = 0
        """:field
        Minimum z positional coordinate of the scene.
        """
        self.z_min: float = 1000
        """:field
        Maximum z positional coordinate of the scene.
        """
        self.z_max: float = 0
        """:field
        All of the rooms in the scene.
        """
        self.rooms: List[RoomBounds] = list()
        for i in range(env.get_num()):
            e = RoomBounds(env=env, i=i)
            if e.x_min < self.x_min:
                self.x_min = e.x_min
            if e.y_min < self.y_min:
                self.y_min = e.y_min
            if e.z_min < self.z_min:
                self.z_min = e.z_min
            if e.x_max > self.x_max:
                self.x_max = e.x_max
            if e.y_max > self.y_max:
                self.y_max = e.y_max
            if e.z_max > self.z_max:
                self.z_max = e.z_max
            self.rooms.append(e)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class Version(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsVersion(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = Version()
        x.Init(buf, n + offset)
        return x

    # Version
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # Version
    def Unity(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Version
    def Tdw(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Version
    def Standalone(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(self._tab.Get(tdw.flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def VersionStart(builder): builder.StartObject(3)
def VersionAddUnity(builder, unity): builder.PrependUOffsetTRelativeSlot(0, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(unity), 0)
def VersionAddTdw(builder, tdw): builder.PrependUOffsetTRelativeSlot(1, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(tdw), 0)
def VersionAddStandalone(builder, standalone): builder.PrependBoolSlot(2, standalone, 0)
def VersionEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class Environment(object):
    __slots__ = ['_tab']

    # Environment
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # Environment
    def Id(self): return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, self._tab.Pos + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(0))
    # Environment
    def Center(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 4)
        return obj

    # Environment
    def Bounds(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 16)
        return obj


def CreateEnvironment(builder, id, center_x, center_y, center_z, bounds_x, bounds_y, bounds_z):
    builder.Prep(4, 28)
    builder.Prep(4, 12)
    builder.PrependFloat32(bounds_z)
    builder.PrependFloat32(bounds_y)
    builder.PrependFloat32(bounds_x)
    builder.Prep(4, 12)
    builder.PrependFloat32(center_z)
    builder.PrependFloat32(center_y)
    builder.PrependFloat32(center_x)
    builder.PrependInt32(id)
    return builder.Offset()



from os import getcwd, chdir
from subprocess import call
from requests import get, head
from typing import Tuple
from platform import system
from pathlib import Path
from zipfile import ZipFile
from distutils import dir_util
import tarfile
from tdw.version import __version__
from tdw.backend.platforms import SYSTEM_TO_RELEASE, SYSTEM_TO_EXECUTABLE


class Build:
    """
    Various helper functions for TDW builds.
    """

    BUILD_ROOT_DIR = Path.home().joinpath(f"tdw_build")
    BUILD_PATH = BUILD_ROOT_DIR.joinpath(f"TDW/TDW{SYSTEM_TO_EXECUTABLE[system()]}")
    if system() == "Darwin":
        BUILD_PATH = BUILD_PATH.joinpath("Contents/MacOS/TDW")

    @staticmethod
    def get_url(version: str = __version__) -> Tuple[str, bool]:
        """
        :param version: The version of the build. Default = the installed version of TDW.

        :return: The URL of the build release matching the version and the OS of this machine, True if the URL exists.
        """

        url = f"https://github.com/threedworld-mit/tdw/releases/download/{version}/" \
              f"{SYSTEM_TO_RELEASE[system()]}"
        if system() == "Windows":
            url += ".zip"
        else:
            url += ".tar.gz"
        # Check if the URL exists.
        if head(url).status_code != 302:
            print(f"Release not found: {url}")
            release_exists = False
        else:
            release_exists = True
        return url, release_exists

    @staticmethod
    def download(version: str = __version__, v_prefix: bool = True) -> bool:
        """
        Download the release corresponding to this version. Move it to the build path and extract it.

        :param version: The version of the build. Default = the installed version of TDW.
        :param v_prefix: If True, add a `v` to the start of the `version` string.

        :return: True if the build downloaded.
        """

        if v_prefix:
            version = "v" + version

        url, url_exists = Build.get_url(version)
        if not url_exists:
            return False

        if Build.BUILD_ROOT_DIR.exists():
            dir_util.remove_tree(str(Build.BUILD_ROOT_DIR.resolve()))
            print("Deleted old release.")

        # Download the build.
        resp = get(url).content
        print("Downloaded the build.")
        # Save the zip file.
        platform = system()
        filename = f"{SYSTEM_TO_RELEASE[platform]}"
        if platform == "Windows":
            filename += ".zip"
        else:
            filename += ".tar.gz"
        zip_path = Path().home().joinpath(filename)
        zip_path.write_bytes(resp)
        print("Saved the file.")

        dst = str(Build.BUILD_ROOT_DIR.resolve())
        # Extract the zip file.
        if platform == "Windows":
            with ZipFile(str(zip_path.resolve()), 'r') as zip_ref:
                zip_ref.extractall(dst)
        else:
            tar = tarfile.open(str(zip_path.resolve()))
            tar.extractall(dst)
            tar.close()
        # Run this to fixed "Damaged App" errors.
        # Source: https://www.google.com/search?client=firefox-b-1-d&q=unity+damaged+app
        if platform == "Darwin":
            cwd = getcwd()
            chdir(str(Build.BUILD_ROOT_DIR.joinpath("TDW").resolve()))
            call(["xattr", "-r", "-d", "com.apple.quarantine", "TDW.app"])
            chdir(cwd)
        print(f"Extracted the file to: {dst}")
        # Delete the zip file.
        zip_path.unlink()
        print("Deleted the download file.")
        return True



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class NavMeshPath(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsNavMeshPath(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = NavMeshPath()
        x.Init(buf, n + offset)
        return x

    # NavMeshPath
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # NavMeshPath
    def Path(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(tdw.flatbuffers.number_types.Uint8Flags, a + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j * 1))
        return 0

    # NavMeshPath
    def PathAsNumpy(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(tdw.flatbuffers.number_types.Uint8Flags, o)
        return 0

    # NavMeshPath
    def PathLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # NavMeshPath
    def State(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 1

    # NavMeshPath
    def Id(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

def NavMeshPathStart(builder): builder.StartObject(3)
def NavMeshPathAddPath(builder, path): builder.PrependUOffsetTRelativeSlot(0, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(path), 0)
def NavMeshPathStartPathVector(builder, numElems): return builder.StartVector(1, numElems, 1)
def NavMeshPathAddState(builder, state): builder.PrependUint8Slot(1, state, 1)
def NavMeshPathAddId(builder, id): builder.PrependInt32Slot(2, id, 0)
def NavMeshPathEnd(builder): return builder.EndObject()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class ContactPoint(object):
    __slots__ = ['_tab']

    # ContactPoint
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # ContactPoint
    def Normal(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 0)
        return obj

    # ContactPoint
    def Point(self, obj):
        obj.Init(self._tab.Bytes, self._tab.Pos + 12)
        return obj


def CreateContactPoint(builder, normal_x, normal_y, normal_z, point_x, point_y, point_z):
    builder.Prep(4, 24)
    builder.Prep(4, 12)
    builder.PrependFloat32(point_z)
    builder.PrependFloat32(point_y)
    builder.PrependFloat32(point_x)
    builder.Prep(4, 12)
    builder.PrependFloat32(normal_z)
    builder.PrependFloat32(normal_y)
    builder.PrependFloat32(normal_x)
    return builder.Offset()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: FBOutput

import tdw.flatbuffers

class Overlap(object):
    __slots__ = ['_tab']

    @classmethod
    def GetRootAsOverlap(cls, buf, offset):
        n = tdw.flatbuffers.encode.Get(tdw.flatbuffers.packer.uoffset, buf, offset)
        x = Overlap()
        x.Init(buf, n + offset)
        return x

    # Overlap
    def Init(self, buf, pos):
        self._tab = tdw.flatbuffers.table.Table(buf, pos)

    # Overlap
    def Id(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, o + self._tab.Pos)
        return 0

    # Overlap
    def ObjectIds(self, j):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(tdw.flatbuffers.number_types.Int32Flags, a + tdw.flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return 0

    # Overlap
    def ObjectIdsAsNumpy(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.GetVectorAsNumpy(tdw.flatbuffers.number_types.Int32Flags, o)
        return 0

    # Overlap
    def ObjectIdsLength(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Overlap
    def Env(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(self._tab.Get(tdw.flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Overlap
    def Walls(self):
        o = tdw.flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(self._tab.Get(tdw.flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

def OverlapStart(builder): builder.StartObject(4)
def OverlapAddId(builder, id): builder.PrependInt32Slot(0, id, 0)
def OverlapAddObjectIds(builder, objectIds): builder.PrependUOffsetTRelativeSlot(1, tdw.flatbuffers.number_types.UOffsetTFlags.py_type(objectIds), 0)
def OverlapStartObjectIdsVector(builder, numElems): return builder.StartVector(4, numElems, 4)
def OverlapAddEnv(builder, env): builder.PrependBoolSlot(2, env, 0)
def OverlapAddWalls(builder, walls): builder.PrependBoolSlot(3, walls, 0)
def OverlapEnd(builder): return builder.EndObject()



from tdw.output_data import Environments


class RoomBounds:
    """
    Data for the bounds of a room in a scene.
    """

    def __init__(self, env: Environments, i: int):
        """
        :param env: The environments output data.
        :param i: The index of this scene in env.get_num()
        """

        """:field
        The ID of the room.
        """
        self.room_id: int = env.get_id(i)
        """:field
        The center of the room.
        """
        self.center = env.get_center(i)
        """:field
        The bounds of the room.
        """
        self.bounds = env.get_bounds(i)
        """:field
        Minimum x positional coordinate of the room.
        """
        self.x_min: float = self.center[0] - (self.bounds[0] / 2)
        """:field
        Minimum y positional coordinate of the room.
        """
        self.y_min: float = self.center[1] - (self.bounds[1] / 2)
        """:field
        Minimum z positional coordinate of the room.
        """
        self.z_min: float = self.center[2] - (self.bounds[2] / 2)
        """:field
        Maximum x positional coordinate of the room.
        """
        self.x_max: float = self.center[0] + (self.bounds[0] / 2)
        """:field
        Maximum y positional coordinate of the room.
        """
        self.y_max: float = self.center[1] + (self.bounds[1] / 2)
        """:field
        Maximum z positional coordinate of the room.
        """
        self.z_max: float = self.center[2] + (self.bounds[2] / 2)

    def is_inside(self, x: float, z: float) -> bool:
        """
        :param x: The x coordinate.
        :param z: The z coordinate.

        :return: True if position (x, z) is in the scene.
        """

        return self.x_min <= x <= self.x_max and self.z_min <= z <= self.z_max


from setuptools import setup, find_packages
from pathlib import Path

__version__ = "1.8.21.0"
readme_path = Path('../README.md')
if readme_path.exists():
    long_description = readme_path.read_text(encoding='utf-8')
else:
    long_description = "See: https://github.com/threedworld-mit/tdw"

setup(
    name='tdw',
    version=__version__,
    description='3D simulation environment',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/threedworld-mit/tdw',
    author_email='alters@mit.edu',
    author='Massachusetts Institute of Technology',
    license='BSD',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    packages=find_packages(),
    include_package_data=True,
    keywords='unity simulation ml machine-learning',
    install_requires=['pyzmq', 'numpy', 'scipy', 'pillow', 'tqdm', 'psutil', 'boto3', 'botocore', 'requests',
                      'pyinstaller'],
)



from typing import Dict, List, Tuple
from tdw.tdw_utils import TDWUtils
from tdw.controller import Controller
from tdw.librarian import ModelLibrarian, ModelRecord
from tdw.py_impact import AudioMaterial, PyImpact, ObjectInfo


class TransformInitData:
    """
    Basic initialization parameters for an object. Can be converted to and from a list of commands.

    This is similar to [`Controller.get_add_object()`](controller.md) except that it includes more parameters.
    """

    LIBRARIES: Dict[str, ModelLibrarian] = dict()
    for _lib_file in ModelLibrarian.get_library_filenames():
        LIBRARIES[_lib_file] = ModelLibrarian(_lib_file)

    def __init__(self, name: str, library: str = "models_core.json", scale_factor: Dict[str, float] = None, position: Dict[str, float] = None, rotation: Dict[str, float] = None, kinematic: bool = False, gravity: bool = True):
        """
        :param name: The name of the model.
        :param library: The filename of the library containing the model's record.
        :param scale_factor: The [scale factor](../api/command_api.md#scale_object).
        :param position: The initial position. If None, defaults to: `{"x": 0, "y": 0, "z": 0`}.
        :param rotation: The initial rotation as Euler angles or a quaternion. If None, defaults to: `{"w": 1, "x": 0, "y": 0, "z": 0}`
        :param kinematic: If True, the object will be [kinematic](../api/command_api.md#set_kinematic_state).
        :param gravity: If True, the object won't respond to [gravity](../api/command_api.md#set_kinematic_state).
        """

        if position is None:
            self.position = TDWUtils.VECTOR3_ZERO
        else:
            self.position = position
        if rotation is None:
            self.rotation = {"w": 1, "x": 0, "y": 0, "z": 0}
        else:
            self.rotation = rotation
        if scale_factor is None:
            self.scale_factor = {"x": 1, "y": 1, "z": 1}
        else:
            self.scale_factor = scale_factor
        self.name = name
        self.library = library
        self.kinematic = kinematic
        self.gravity = gravity

    def get_commands(self) -> Tuple[int, List[dict]]:
        """
        :return: Tuple: The ID of the object; a list of commands to create the object: `[add_object, rotate_object_to, scale_object, set_kinematic_state, set_object_collision_detection_mode]`
        """

        record = self._get_record()

        object_id = Controller.get_unique_id()
        commands = [{"$type": "add_object",
                     "name": record.name,
                     "url": record.get_url(),
                     "scale_factor": record.scale_factor,
                     "position": self.position,
                     "category": record.wcategory,
                     "id": object_id}]
        # The rotation is a quaternion.
        if "w" in self.rotation:
            commands.append({"$type": "rotate_object_to",
                             "rotation": self.rotation,
                             "id": object_id})
        # The rotation is in Euler angles.
        else:
            commands.append({"$type": "rotate_object_to_euler_angles",
                             "euler_angles": self.rotation,
                             "id": object_id})
        commands.extend([{"$type": "scale_object",
                          "scale_factor": self.scale_factor,
                          "id": object_id},
                         {"$type": "set_kinematic_state",
                          "id": object_id,
                          "is_kinematic": self.kinematic,
                          "use_gravity": self.gravity}])
        # Kinematic objects must be continuous_speculative.
        if self.kinematic:
            commands.append({"$type": "set_object_collision_detection_mode",
                             "id": object_id,
                             "mode": "continuous_speculative"})

        return object_id, commands

    def _get_record(self) -> ModelRecord:
        """
        :return: The model metadata record for this object.
        """

        return TransformInitData.LIBRARIES[self.library].get_record(name=self.name)


class RigidbodyInitData(TransformInitData):
    """
    A subclass of `TransformInitData`. Includes data and commands to set the mass and physic material of the object.
    """

    def __init__(self, name: str, mass: float, dynamic_friction: float, static_friction: float, bounciness: float, library: str = "models_core.json", scale_factor: Dict[str, float] = None, position: Dict[str, float] = None, rotation: Dict[str, float] = None, kinematic: bool = False, gravity: bool = True):
        """
        :param name: The name of the model.
        :param library: The filename of the library containing the model's record.
        :param scale_factor: The [scale factor](../api/command_api.md#scale_object).
        :param position: The initial position. If None, defaults to: `{"x": 0, "y": 0, "z": 0`}.
        :param rotation: The initial rotation as Euler angles or a quaternion. If None, defaults to: `{"w": 1, "x": 0, "y": 0, "z": 0}`
        :param kinematic: If True, the object will be [kinematic](../api/command_api.md#set_kinematic_state).
        :param gravity: If True, the object won't respond to [gravity](../api/command_api.md#set_kinematic_state).
        :param mass: The mass of the object.
        :param dynamic_friction: The [dynamic friction](../api/command_api.md#set_physic_material) of the object.
        """

        super().__init__(name=name, library=library, scale_factor=scale_factor, position=position, rotation=rotation,
                         kinematic=kinematic, gravity=gravity)
        self.mass = mass
        self.dynamic_friction = dynamic_friction
        self.static_friction = static_friction
        self.bounciness = bounciness

    def get_commands(self) -> Tuple[int, List[dict]]:
        """
        :return: Tuple: The ID of the object; a list of commands to create the object: `[add_object, rotate_object_to, scale_object, set_kinematic_state, set_object_collision_detection_mode, set_mass, set_physic_material]`
        """

        object_id, commands = super().get_commands()
        # Set the mass and physic material.
        commands.extend([{"$type": "set_mass",
                          "mass": self.mass,
                          "id": object_id},
                         {"$type": "set_physic_material",
                          "dynamic_friction": self.dynamic_friction,
                          "static_friction": self.static_friction,
                          "bounciness": self.bounciness,
                          "id": object_id}])
        return object_id, commands


class AudioInitData(RigidbodyInitData):
    """
    A subclass of `RigidbodyInitData` that includes [audio values](py_impact.md#objectinfo).
    Physics values are derived from these audio values.
    """

    _DYNAMIC_FRICTION = {AudioMaterial.ceramic: 0.47,
                         AudioMaterial.wood_hard: 0.35,
                         AudioMaterial.wood_medium: 0.35,
                         AudioMaterial.wood_soft: 0.35,
                         AudioMaterial.cardboard: 0.45,
                         AudioMaterial.paper: 0.47,
                         AudioMaterial.glass: 0.65,
                         AudioMaterial.fabric: 0.65,
                         AudioMaterial.leather: 0.4,
                         AudioMaterial.stone: 0.7,
                         AudioMaterial.rubber: 0.75,
                         AudioMaterial.plastic_hard: 0.3,
                         AudioMaterial.plastic_soft_foam: 0.45,					 
                         AudioMaterial.metal: 0.43}
    _STATIC_FRICTION = {AudioMaterial.ceramic: 0.47,
                        AudioMaterial.wood_hard: 0.37,
                        AudioMaterial.wood_medium: 0.37,
                        AudioMaterial.wood_soft: 0.37,
                        AudioMaterial.cardboard: 0.48,
                        AudioMaterial.paper: 0.5,
                        AudioMaterial.glass: 0.68,
                        AudioMaterial.fabric: 0.67,
                        AudioMaterial.leather: 0.43,
                        AudioMaterial.stone: 0.72,
                        AudioMaterial.rubber: 0.8,
                        AudioMaterial.plastic_hard: 0.35,
                        AudioMaterial.plastic_soft_foam: 0.47,
                        AudioMaterial.metal: 0.47}
    AUDIO = PyImpact.get_object_info()

    def __init__(self, name: str, library: str = "models_core.json", scale_factor: Dict[str, float] = None, position: Dict[str, float] = None, rotation: Dict[str, float] = None, kinematic: bool = False, gravity: bool = True, audio: ObjectInfo = None):
        """
        :param name: The name of the model.
        :param library: The filename of the library containing the model's record.
        :param scale_factor: The [scale factor](../api/command_api.md#scale_object).
        :param position: The initial position. If None, defaults to: `{"x": 0, "y": 0, "z": 0`}.
        :param rotation: The initial rotation as Euler angles or a quaternion. If None, defaults to: `{"w": 1, "x": 0, "y": 0, "z": 0}`
        :param kinematic: If True, the object will be [kinematic](../api/command_api.md#set_kinematic_state).
        :param gravity: If True, the object won't respond to [gravity](../api/command_api.md#set_kinematic_state).
        :param audio: If None, derive physics data from the audio data in `PyImpact.get_object_info()` (if the object isn't in this dictionary, this constructor will throw an error). If not None, use these values instead of the default audio values.
        """

        if audio is None:
            self.audio = AudioInitData.AUDIO[name]
        else:
            self.audio = audio
        super().__init__(name=name, library=library, scale_factor=scale_factor, position=position, rotation=rotation,
                         kinematic=kinematic, gravity=gravity, mass=self.audio.mass,
                         dynamic_friction=AudioInitData._DYNAMIC_FRICTION[self.audio.material],
                         static_friction=AudioInitData._STATIC_FRICTION[self.audio.material],
                         bounciness=self.audio.bounciness)

    def get_commands(self) -> Tuple[int, List[dict]]:
        """
        :return: Tuple: The ID of the object; a list of commands to create the object: `[add_object, rotate_object_to, scale_object, set_kinematic_state, set_object_collision_detection_mode, set_mass, set_physic_material]`
        """

        return super().get_commands()
